import os
import sys
# In order to use connectionGlobals we need to adjust the sys path
PROJECT_ROOT = os.path.abspath(os.path.join(
                  os.path.dirname(__file__),
                  os.pardir)
)
sys.path.clear()
sys.path.append(PROJECT_ROOT)
from RSTAB.initModel import Model
from RSTAB import connectionGlobals

def importFrom(targetFilePath: str):
    '''
    Allowed file extensions are .xml, .saf and .xlsx.

    Args:
        targetFilePath (string): Destination path to the file
    '''
    connectionGlobals.client.service.import_from(targetFilePath)
    head, tail = os.path.split(targetFilePath)
    if '.' in tail:
            tail = tail.split('.')[0]
    Model(False, tail)

def getConversionTables():
    '''
    Get conversion tables.
    '''
    return connectionGlobals.client.service.get_conversion_tables()

def setConversionTables(ConversionTables):
    '''
    Set conversion tables.

    Args:
        ConversionTables (ns0:ConversionTables): Conversion tables structure
    '''
    connectionGlobals.client.service.set_conversion_tables(ConversionTables)

def getSAFSettings():
    '''
    Get SAF import/export settings.
    '''
    return connectionGlobals.client.service.get_saf_settings()

def setSAFSettings(SafConfiguration):
    '''
    Set SAF import/export settings.

    Args:
        SafConfiguration (ns0:SafConfiguration) SAF settings obtained by getSAFSettings()
    '''
    connectionGlobals.client.service.set_saf_settings(SafConfiguration)


from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString
from RSTAB.enums import *

class SteelEffectiveLengths():
    def __init__(self,
                 no: int = 1,
                 members: str = '1',
                 member_sets: str = '',
                 flexural_buckling_about_y: bool = True,
                 flexural_buckling_about_z: bool = True,
                 torsional_buckling: bool = True,
                 lateral_torsional_buckling: bool = True,
                 principal_section_axes: bool = True,
                 geometric_section_axes: bool = True,
                 name: str = 'SEL1',
                 nodal_supports: list = [
                     [SteelEffectiveLengthsSupportType.SUPPORT_TYPE_FIXED_IN_Z_Y_AND_TORSION, True, 0.0, SteelEffectiveLengthsEccentricityType.ECCENTRICITY_TYPE_NONE, \
                      0.0, 0.0, 0.0, 0.0, SteelEffectiveLengthsSupportTypeInY.SUPPORT_STATUS_YES, SteelEffectiveLengthsRestraintTypeAboutX.SUPPORT_STATUS_YES, \
                      SteelEffectiveLengthsRestraintTypeAboutZ.SUPPORT_STATUS_NO, SteelEffectiveLengthsRestraintTypeWarping.SUPPORT_STATUS_NO, "1"],
                     [SteelEffectiveLengthsSupportType.SUPPORT_TYPE_FIXED_IN_Z_Y_AND_TORSION, True, 0.0, SteelEffectiveLengthsEccentricityType.ECCENTRICITY_TYPE_NONE, \
                      0.0, 0.0, 0.0, 0.0, SteelEffectiveLengthsSupportTypeInY.SUPPORT_STATUS_YES, SteelEffectiveLengthsRestraintTypeAboutX.SUPPORT_STATUS_YES, \
                      SteelEffectiveLengthsRestraintTypeAboutZ.SUPPORT_STATUS_NO, SteelEffectiveLengthsRestraintTypeWarping.SUPPORT_STATUS_NO, "2"]
                                        ],
                 factors: list = [
                     [1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                                 ],
                 intermediate_nodes: bool = False,
                 different_properties: bool = True,
                 factors_definition_absolute: bool = False,
                 import_from_stability_analysis_enabled: bool = False,
                 determination_of_mcr = SteelEffectiveLengthsDeterminationMcrEurope.DETERMINATION_EUROPE_EIGENVALUE,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Steel Effective Length Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            flexural_buckling_about_y (bool): Flexural Buckling About Y Option
            flexural_buckling_about_z (bool): Flexural Buckling About Z Option
            torsional_buckling (bool): Torsional Buckling Option
            lateral_torsional_buckling (bool): Lateral Torsional Buckling Option
            principal_section_axes (bool): Principal Section Axes Option
            geometric_section_axes (bool): Geometric Section Axes Option
            name (str): User Defined Effective Length Name
            nodal_supports (lst): Nodal Support Table Definition
                nodal_supports[i][0] (enum): Support Type Enumeration Type
                nodal_supports[i][1] (bool): Support in Z Option
                nodal_supports[i][2] (float): Support Spring in Y Coefficient
                nodal_supports[i][3] (enum): Eccentricity Type Enumeration Type
                nodal_supports[i][4] (float): Eccentricity in Z Direction
                nodal_supports[i][5] (float): Restraint Spring About X Coefficient
                nodal_supports[i][6] (float): Restraint Spring About Z Coefficient
                nodal_supports[i][7] (float): Restraint Spring Warping Coefficient
                nodal_supports[i][8] (enum): Support Type in Y Enumeration Type
                nodal_supports[i][9] (enum): Restraint Type in X Enumeration Type
                nodal_supports[i][10] (enum): Restraint Type in Z Enumeration Type
                nodal_supports[i][11] (enum): Restraint Type Warping Enumeration Type
                nodal_supports[i][12] (str): Assigned Nodes
            factors (list of lists): Effective Length Factors
                factors[i][0] (float): Flexural Buckling in U Coefficient
                factors[i][1] (float): Flexural Buckling in V Coefficient
                factors[i][2] (float): Flexural Buckling in Y Coefficient
                factors[i][3] (float): Flexural Buckling in Z Coefficient
                factors[i][4] (float): Torsional Buckling Coefficient
                factors[i][5] (float): Lateral Torsional Buckling Coefficient
                factors[i][6] (float): Lateral Torsional Buckling Top Coefficient
                factors[i][7] (float): Lateral Torsional Buckling Bottom Coefficient
                factors[i][8] (float): Twist Restraint Coefficient
                factors[i][9] (float): Lateral Torsional Restraint Coefficient
                factors[i][10] (float): Critical Moment
            intermediate_nodes (bool): Intermediate Nodes Option
            different_properties (bool): Different Properties Option
            factors_definition_absolute (bool): Absolute Factors Definition Option
            import_from_stability_analysis_enabled (bool): Import From Stability Analysis Option
            determination_of_mcr (enum): Steel Effective Lengths Determination Mcr Europe Enumeration Item
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Types For Steel Design Effective Lengths
        clientObject = model.clientModel.factory.create('ns0:steel_effective_lengths')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Effective Lengths No.
        clientObject.no = no

        # Effective Lengths Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Effective Lengths Assigned Member Sets
        clientObject.member_sets = ConvertToDlString(member_sets)

        # Effective Lengths Flexural Buckling About Y
        clientObject.flexural_buckling_about_y = flexural_buckling_about_y

        # Effective Lengths Flexural Buckling About Z
        clientObject.flexural_buckling_about_z = flexural_buckling_about_z

        # Effective Lengths Torsional Buckling
        clientObject.torsional_buckling = torsional_buckling

        # Effective Lengths Lateral Torsional Buckling
        clientObject.lateral_torsional_buckling = lateral_torsional_buckling

        # Effective Lengths Principal Section Axes
        clientObject.principal_section_axes = principal_section_axes

        # Effective Lengths Geometric Section Axes
        clientObject.geometric_section_axes = geometric_section_axes

        # Effective Lengths User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Effective Lengths Nodal Supports
        clientObject.nodal_supports = model.clientModel.factory.create('ns0:array_of_steel_effective_lengths_nodal_supports')

        for i,j in enumerate(nodal_supports):
            mlvlp = model.clientModel.factory.create('ns0:steel_effective_lengths_nodal_supports_row')
            mlvlp.no = i+1
            mlvlp.row = model.clientModel.factory.create('ns0:steel_effective_lengths_nodal_supports')
            clearAttributes(mlvlp.row)
            mlvlp.row.support_type = nodal_supports[i][0].name
            mlvlp.row.support_in_z = nodal_supports[i][1]
            mlvlp.row.support_spring_in_y = nodal_supports[i][2]
            mlvlp.row.eccentricity_type = nodal_supports[i][3].name
            mlvlp.row.eccentricity_ez = nodal_supports[i][4]
            mlvlp.row.restraint_spring_about_x = nodal_supports[i][5]
            mlvlp.row.restraint_spring_about_z = nodal_supports[i][6]
            mlvlp.row.restraint_spring_warping = nodal_supports[i][7]
            mlvlp.row.support_in_y_type = nodal_supports[i][8].name
            mlvlp.row.restraint_about_x_type = nodal_supports[i][9].name
            mlvlp.row.restraint_about_z_type = nodal_supports[i][10].name
            mlvlp.row.restraint_warping_type = nodal_supports[i][11].name
            mlvlp.row.nodes = ConvertToDlString(nodal_supports[i][12])

            clientObject.nodal_supports.steel_effective_lengths_nodal_supports.append(mlvlp)

        # Effective Lengths Factors
        clientObject.factors = model.clientModel.factory.create('ns0:array_of_steel_effective_lengths_factors')

        for i,j in enumerate(factors):
            mlvlp_f = model.clientModel.factory.create('ns0:steel_effective_lengths_factors_row')
            mlvlp_f.no = i+1
            mlvlp_f.row = model.clientModel.factory.create('ns0:steel_effective_lengths_factors')
            clearAttributes(mlvlp_f.row)
            mlvlp_f.row.flexural_buckling_u = factors[i][0]
            mlvlp_f.row.flexural_buckling_v = factors[i][1]
            mlvlp_f.row.flexural_buckling_y = factors[i][2]
            mlvlp_f.row.flexural_buckling_z = factors[i][3]
            mlvlp_f.row.torsional_buckling = factors[i][4]
            mlvlp_f.row.lateral_torsional_buckling = factors[i][5]
            mlvlp_f.row.lateral_torsional_buckling_top = factors[i][6]
            mlvlp_f.row.lateral_torsional_buckling_bottom = factors[i][7]
            mlvlp_f.row.twist_restraint = factors[i][8]
            mlvlp_f.row.lateral_torsional_restraint = factors[i][9]
            mlvlp_f.row.critical_moment = factors[i][10]

            clientObject.factors.steel_effective_lengths_factors.append(mlvlp_f)

        # Effective Lengths Intermediate Points
        clientObject.intermediate_nodes = intermediate_nodes

        # Effective Lengths Different Properties
        clientObject.different_properties = different_properties

        # Effective Lengths Factors Definition Absolute
        clientObject.factors_definition_absolute = factors_definition_absolute

        # Effective Lengths Import From Stability Analysis Enabled
        clientObject.import_from_stability_analysis_enabled = import_from_stability_analysis_enabled

        # Effective Lengths Determination MCR Europe
        clientObject.determination_mcr_europe = determination_of_mcr.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Steel Effective Lengths to client model
        model.clientModel.service.set_steel_effective_lengths(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import AnalysisType

class LoadCombination():

    def __init__(self,
                 no: int = 1,
                 analysis_type = AnalysisType.ANALYSIS_TYPE_STATIC,
                 design_situation: int = 1,
                 name: str = '',
                 static_analysis_settings: int = 1,
                 consider_imperfection: bool = False,
                 consider_initial_state: bool = False,
                 structure_modification: bool = False,
                 to_solve: bool = True,
                 combination_items = [[1.5, 1, 0, False]],
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Args:
            no (int): Load Combination Tag
            analysis_type (enum): Analysis Type Enumeration
            design_situation (int): Design Situation
            name (str): User defined Combination Name
            static_analysis_settings (int): Static Analysis Settings Number
            consider_imperfection (bool): Consider Imperfection Options
            consider_initial_state (bool): Consider Initial State
            structure_modification (bool): Enable/Disable Structure Modification
            to_solve (bool): Decide to solve
            combination_items (list of list): Combination Items
                for Combination Items;
                    combination_items = [[factor, load_case, action, is_leading],...]
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Load Combination
        clientObject = model.clientModel.factory.create('ns0:load_combination')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Load Combination No.
        clientObject.no = no

        # Analysis Type
        clientObject.analysis_type = analysis_type.name

        # Design Situation Assignment
        clientObject.design_situation = design_situation

        # Combination Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Analysis Settings Assignment
        clientObject.static_analysis_settings = static_analysis_settings

        # Consider Imperfection Options
        clientObject.consider_imperfection = consider_imperfection

        # Consider Initial State
        clientObject.consider_initial_state = consider_initial_state

        # Structure Modification Enable
        clientObject.structure_modification_enabled = structure_modification

        # Decide to Solve
        clientObject.to_solve = to_solve

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Items
        clientObject.items = model.clientModel.factory.create('ns0:load_combination.items')

        for i,j in enumerate(combination_items):
            mlvlp = model.clientModel.factory.create('ns0:load_combination_items_row')
            mlvlp.no = i+1
            mlvlp.row = model.clientModel.factory.create('ns0:load_combination_items')
            clearAttributes(mlvlp.row)
            mlvlp.row.factor = combination_items[i][0]
            mlvlp.row.load_case = combination_items[i][1]
            mlvlp.row.action = combination_items[i][2]
            mlvlp.row.is_leading = combination_items[i][3]

            clientObject.items.load_combination_items.append(mlvlp)

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Load Combination to client model
        model.clientModel.service.set_load_combination(clientObject)



from RSTAB.enums import MemberHingeNonlinearity
from RSTAB.initModel import Model, clearAttributes, ConvertToDlString, deleteEmptyAttributes
from RSTAB.dataTypes import inf

class MemberHinge():
    def __init__(self,
                 no: int = 1,
                 coordinate_system: str = "Local",
                 member : str = "",
                 translational_release_n: float = inf,
                 translational_release_vy: float = inf,
                 translational_release_vz: float = inf,
                 rotational_release_mt: float = inf,
                 rotational_release_my: float = 0.0,
                 rotational_release_mz: float = 0.0,
                 translational_release_n_nonlinearity = [MemberHingeNonlinearity.NONLINEARITY_TYPE_NONE],
                 translational_release_vy_nonlinearity = [MemberHingeNonlinearity.NONLINEARITY_TYPE_NONE],
                 translational_release_vz_nonlinearity = [MemberHingeNonlinearity.NONLINEARITY_TYPE_NONE],
                 rotational_release_mt_nonlinearity = [MemberHingeNonlinearity.NONLINEARITY_TYPE_NONE],
                 rotational_release_my_nonlinearity = [MemberHingeNonlinearity.NONLINEARITY_TYPE_NONE],
                 rotational_release_mz_nonlinearity = [MemberHingeNonlinearity.NONLINEARITY_TYPE_NONE],
                 name: str = None,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Member Hinge Tag
            coordinate_system (str): Coordinate System Selection
            member (str): Assigned Members
            translational_release_n (float): Translational Spring Constant X
            translational_release_vy (float): Translational Spring Constant Y
            translational_release_vz (float): Translational Spring Constant Z
            rotational_release_mt (float): Rotational Spring Constant X
            rotational_release_my (float): Rotational Spring Constant Y
            rotational_release_mz (float): Rotational Spring Constant Z
            translational_release_n_nonlinearity (list): Nonlinearity Options Translational X
            translational_release_vy_nonlinearity (list): Nonlinearity Options Translational Y
            translational_release_vz_nonlinearity (list): Nonlinearity Options Translational Z
                for translational_release_n/vy/vz_nonlinearity[0] == MemberHingeNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Partial_Activity, negative zone, positive zone]
                    for negative/positive zone[0] == MemberHingePartialActivityType.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                        negative/positive zone = [negative/positive zone type, slippage]
                    for negative/positive zone[0] == MemberHingePartialActivityType.PARTIAL_ACTIVITY_TYPE_FIXED:
                        negative/positive zone = [negative/positive zone type, slippage, displacement]  (Note: Displacement must be greater than slippage)
                    for negative/positive zone[0] == MemberHingePartialActivityType.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE/PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE:
                        negative/positive zone = [negative/positive zone type, slippage, force]
                for translational_release_n/vy/vz_nonlinearity[0] == MemberHingeNonlinearity.NONLINEARITY_TYPE_DIAGRAM:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Diagram, [symmetric(bool), MemberHingeDiagram Enumeration(start), MemberHingeDiagram Enumeration(end)], [[displacement, force],...]]
                for translational_release_n/vy/vz_nonlinearity[0] == MemberHingeNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1/NONLINEARITY_TYPE_FRICTION_DIRECTION_2/NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Friction Direction, friction coefficient(float)]
                for translational_release_n/vy/vz_nonlinearity[0] == MemberHingeNonlinearity.NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2:
                    translational_release_n/vy/vz_nonlinearity = [nonlinearity type Friction Direction, friction coefficient 1(float), friction coefficient 2(float)]
            rotational_release_mt_nonlinearity (list): Nonlinearity Options Rotational X
            rotational_release_my_nonlinearity (list): Nonlinearity Options Rotational Y
            rotational_release_mz_nonlinearity (list): Nonlinearity Options Rotational Z
                for rotational_release_mt/my/mz_nonlinearity[0] == MemberHingeNonlinearity.NONLINEARITY_TYPE_PARTIAL_ACTIVITY:
                    rotational_release_mt/my/mz_nonlinearity = [nonlinearity type Partial_Activity, negative zone, positive zone]
                    for negative/positive zone[0] == MemberHingePartialActivityType.PARTIAL_ACTIVITY_TYPE_COMPLETE:
                        negative/positive zone = [negative/positive zone type, slippage]
                    for negative/positive zone[0] == MemberHingePartialActivityType.PARTIAL_ACTIVITY_TYPE_FIXED:
                        negative/positive zone = [negative/positive zone type, slippage, rotation]
                    for negative/positive zone[0] == MemberHingePartialActivityType.PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT/PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT:
                        negative/positive zone = [negative/positive zone type, slippage, moment]
                for rotational_release_mt/my/mz_nonlinearity[0] == MemberHingePartialActivityType.NONLINEARITY_TYPE_DIAGRAM:
                    rotational_release_mt/my/mz_nonlinearity = [nonlinearity type Diagram, [symmetric(bool), NodalReleaseTypeDiagram Enumeration(start), NodalReleaseTypeDiagram Enumeration(end)], [[rotation, moment],...]]
            name (str, optional): User Defined Member Hinge Name
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Member Hinge
        clientObject = model.clientModel.factory.create('ns0:member_hinge')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Hinge No.
        clientObject.no = no

        # Assigned Member
        clientObject.members = ConvertToDlString(member)

        # Coordinate System
        clientObject.coordinate_system = coordinate_system

        # Translational Release/Spring [N/m] N
        clientObject.axial_release_n = translational_release_n

        # Translational Release/Spring [N/m] Vy
        clientObject.axial_release_vy = translational_release_vy

        # Translational Release/Spring [N/m] Vz
        clientObject.axial_release_vz = translational_release_vz

        # Rotational Release/Spring [Nm/rad] Mt
        clientObject.moment_release_mt = rotational_release_mt

        # Rotational Release/Spring [Nm/rad] My
        clientObject.moment_release_my = rotational_release_my

        # Rotational Release/Spring [Nm/rad] Mz
        clientObject.moment_release_mz = rotational_release_mz

        # Translational Release N Nonlinearity

        # Nonlinearity Types None, Fixed if Negative N, Fixed if Positive N
        if translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_NONE" \
        or translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE" \
        or translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_POSITIVE" :
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name

        # Partial Activity
        elif translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_PARTIAL_ACTIVITY":
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_along_x_negative_type = translational_release_n_nonlinearity[1][0].name

            if translational_release_n_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][1]

            elif translational_release_n_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][1]
                clientObject.partial_activity_along_x_negative_displacement = translational_release_n_nonlinearity[1][2]

            elif translational_release_n_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE":
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][1]
                clientObject.partial_activity_along_x_negative_force = translational_release_n_nonlinearity[1][2]

            elif translational_release_n_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE":
                clientObject.partial_activity_along_x_negative_slippage = translational_release_n_nonlinearity[1][1]
                clientObject.partial_activity_along_x_negative_force = translational_release_n_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_along_x_positive_type = translational_release_n_nonlinearity[2][0].name

            if translational_release_n_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][1]

            elif translational_release_n_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][1]
                clientObject.partial_activity_along_x_positive_displacement = translational_release_n_nonlinearity[2][2]

            elif translational_release_n_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE":
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][1]
                clientObject.partial_activity_along_x_positive_force = translational_release_n_nonlinearity[2][2]

            elif translational_release_n_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE":
                clientObject.partial_activity_along_x_positive_slippage = translational_release_n_nonlinearity[2][1]
                clientObject.partial_activity_along_x_positive_force = translational_release_n_nonlinearity[2][2]

        elif translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_DIAGRAM":
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name
            clientObject.diagram_along_x_symmetric = translational_release_n_nonlinearity[1][0]
            clientObject.diagram_along_x_is_sorted = True

            if translational_release_n_nonlinearity[1][0]:
                clientObject.diagram_along_x_start = translational_release_n_nonlinearity[1][1].name
                clientObject.diagram_along_x_end = translational_release_n_nonlinearity[1][1].name

            else:
                clientObject.diagram_along_x_start = translational_release_n_nonlinearity[1][1].name
                clientObject.diagram_along_x_end = translational_release_n_nonlinearity[1][2].name

            clientObject.diagram_along_x_table = model.clientModel.factory.create('ns0:member_hinge.diagram_along_x_table')

            for i,j in enumerate(translational_release_n_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:member_hinge_diagram_along_x_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:member_hinge_diagram_along_x_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.displacement = translational_release_n_nonlinearity[2][i][0]
                mlvlp.row.force = translational_release_n_nonlinearity[2][i][1]

                clientObject.diagram_along_x_table.member_hinge_diagram_along_x_table.append(mlvlp)

        elif translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1" \
        or translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_2" \
        or translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2":
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name
            clientObject.friction_coefficient_x = translational_release_n_nonlinearity[1]

        elif translational_release_n_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2":
            clientObject.axial_release_n_nonlinearity = translational_release_n_nonlinearity[0].name
            clientObject.friction_coefficient_xy = translational_release_n_nonlinearity[1]
            clientObject.friction_coefficient_xz = translational_release_n_nonlinearity[2]

        # Translational Release Vy Nonlinearity
        if translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_NONE" \
        or translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE" \
        or translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_POSITIVE" :
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name

        # Partial Activity
        elif translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_PARTIAL_ACTIVITY":
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_along_y_negative_type = translational_release_vy_nonlinearity[1][0].name

            if translational_release_vy_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][1]

            elif translational_release_vy_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][1]
                clientObject.partial_activity_along_y_negative_displacement = translational_release_vy_nonlinearity[1][2]

            elif translational_release_vy_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE":
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][1]
                clientObject.partial_activity_along_y_negative_force = translational_release_vy_nonlinearity[1][2]

            elif translational_release_vy_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE":
                clientObject.partial_activity_along_y_negative_slippage = translational_release_vy_nonlinearity[1][1]
                clientObject.partial_activity_along_y_negative_force = translational_release_vy_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_along_y_positive_type = translational_release_vy_nonlinearity[2][0].name

            if translational_release_vy_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][1]

            elif translational_release_vy_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][1]
                clientObject.partial_activity_along_y_positive_displacement = translational_release_vy_nonlinearity[2][2]

            elif translational_release_vy_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE":
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][1]
                clientObject.partial_activity_along_y_positive_force = translational_release_vy_nonlinearity[2][2]

            elif translational_release_vy_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE":
                clientObject.partial_activity_along_y_positive_slippage = translational_release_vy_nonlinearity[2][1]
                clientObject.partial_activity_along_y_positive_force = translational_release_vy_nonlinearity[2][2]

        elif translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_DIAGRAM":
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name
            clientObject.diagram_along_y_symmetric = translational_release_vy_nonlinearity[1][0]
            clientObject.diagram_along_y_is_sorted = True

            if translational_release_vy_nonlinearity[1][0]:
                clientObject.diagram_along_y_start = translational_release_vy_nonlinearity[1][1].name
                clientObject.diagram_along_y_end = translational_release_vy_nonlinearity[1][1].name

            else:
                clientObject.diagram_along_y_start = translational_release_vy_nonlinearity[1][1].name
                clientObject.diagram_along_y_end = translational_release_vy_nonlinearity[1][2].name

            clientObject.diagram_along_y_table = model.clientModel.factory.create('ns0:member_hinge.diagram_along_y_table')

            for i,j in enumerate(translational_release_vy_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:member_hinge_diagram_along_y_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:member_hinge_diagram_along_y_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.displacement = translational_release_vy_nonlinearity[2][i][0]
                mlvlp.row.force = translational_release_vy_nonlinearity[2][i][1]

                clientObject.diagram_along_y_table.member_hinge_diagram_along_y_table.append(mlvlp)

        elif translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1" \
        or translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_2" \
        or translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2":
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name
            clientObject.friction_coefficient_y = translational_release_vy_nonlinearity[1]

        elif translational_release_vy_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2":
            clientObject.axial_release_vy_nonlinearity = translational_release_vy_nonlinearity[0].name
            clientObject.friction_coefficient_yx = translational_release_vy_nonlinearity[1]
            clientObject.friction_coefficient_yz = translational_release_vy_nonlinearity[2]

        # Translational Release Vz Nonlinearity
        if translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_NONE" \
        or translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE" \
        or translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_POSITIVE" :
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name

        # Partial Activity
        elif translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_PARTIAL_ACTIVITY":
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_along_z_negative_type = translational_release_vz_nonlinearity[1][0].name

            if translational_release_vz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][1]

            elif translational_release_vz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][1]
                clientObject.partial_activity_along_z_negative_displacement = translational_release_vz_nonlinearity[1][2]

            elif translational_release_vz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE":
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][1]
                clientObject.partial_activity_along_z_negative_force = translational_release_vz_nonlinearity[1][2]

            elif translational_release_vz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE":
                clientObject.partial_activity_along_z_negative_slippage = translational_release_vz_nonlinearity[1][1]
                clientObject.partial_activity_along_z_negative_force = translational_release_vz_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_along_z_positive_type = translational_release_vz_nonlinearity[2][0].name

            if translational_release_vz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][1]

            elif translational_release_vz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][1]
                clientObject.partial_activity_along_z_positive_displacement = translational_release_vz_nonlinearity[2][2]

            elif translational_release_vz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_FORCE":
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][1]
                clientObject.partial_activity_along_z_positive_force = translational_release_vz_nonlinearity[2][2]

            elif translational_release_vz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_FORCE":
                clientObject.partial_activity_along_z_positive_slippage = translational_release_vz_nonlinearity[2][1]
                clientObject.partial_activity_along_z_positive_force = translational_release_vz_nonlinearity[2][2]

        elif translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_DIAGRAM":
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name
            clientObject.diagram_along_z_symmetric = translational_release_vz_nonlinearity[1][0]
            clientObject.diagram_along_z_is_sorted = True

            if translational_release_vz_nonlinearity[1][0]:
                clientObject.diagram_along_z_start = translational_release_vz_nonlinearity[1][1].name
                clientObject.diagram_along_z_end = translational_release_vz_nonlinearity[1][1].name

            else:
                clientObject.diagram_along_z_start = translational_release_vz_nonlinearity[1][1].name
                clientObject.diagram_along_z_end = translational_release_vz_nonlinearity[1][2].name

            clientObject.diagram_along_z_table = model.clientModel.factory.create('ns0:member_hinge.diagram_along_z_table')

            for i,j in enumerate(translational_release_vz_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:member_hinge_diagram_along_z_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:member_hinge_diagram_along_z_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.displacement = translational_release_vz_nonlinearity[2][i][0]
                mlvlp.row.force = translational_release_vz_nonlinearity[2][i][1]

                clientObject.diagram_along_z_table.member_hinge_diagram_along_z_table.append(mlvlp)

        elif translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1" \
        or translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_2" \
        or translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1_2":
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name
            clientObject.friction_coefficient_z = translational_release_vz_nonlinearity[1]

        elif translational_release_vz_nonlinearity[0].name == "NONLINEARITY_TYPE_FRICTION_DIRECTION_1_PLUS_2":
            clientObject.axial_release_vz_nonlinearity = translational_release_vz_nonlinearity[0].name
            clientObject.friction_coefficient_zx = translational_release_vz_nonlinearity[1]
            clientObject.friction_coefficient_zy = translational_release_vz_nonlinearity[2]

        # Rotational Release Mt Nonlinearity
        if rotational_release_mt_nonlinearity[0].name == "NONLINEARITY_TYPE_NONE" \
        or rotational_release_mt_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE" \
        or rotational_release_mt_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_POSITIVE" :
            clientObject.moment_release_mt_nonlinearity = rotational_release_mt_nonlinearity[0].name

        # Partial Activity
        elif rotational_release_mt_nonlinearity[0].name == "NONLINEARITY_TYPE_PARTIAL_ACTIVITY":
            clientObject.moment_release_mt_nonlinearity = rotational_release_mt_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_around_x_negative_type = rotational_release_mt_nonlinearity[1][0].name

            if rotational_release_mt_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][1]

            elif rotational_release_mt_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][1]
                clientObject.partial_activity_around_x_negative_rotation = rotational_release_mt_nonlinearity[1][2]

            elif rotational_release_mt_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT":
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][1]
                clientObject.partial_activity_around_x_negative_moment = rotational_release_mt_nonlinearity[1][2]

            elif rotational_release_mt_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT":
                clientObject.partial_activity_around_x_negative_slippage = rotational_release_mt_nonlinearity[1][1]
                clientObject.partial_activity_around_x_negative_force = rotational_release_mt_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_around_x_positive_type = rotational_release_mt_nonlinearity[2][0].name

            if rotational_release_mt_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][1]

            elif rotational_release_mt_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][1]
                clientObject.partial_activity_around_x_positive_rotation = rotational_release_mt_nonlinearity[2][2]

            elif rotational_release_mt_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT":
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][1]
                clientObject.partial_activity_around_x_positive_moment = rotational_release_mt_nonlinearity[2][2]

            elif rotational_release_mt_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT":
                clientObject.partial_activity_around_x_positive_slippage = rotational_release_mt_nonlinearity[2][1]
                clientObject.partial_activity_around_x_positive_moment = rotational_release_mt_nonlinearity[2][2]

        elif rotational_release_mt_nonlinearity[0].name == "NONLINEARITY_TYPE_DIAGRAM":
            clientObject.moment_release_mt_nonlinearity = rotational_release_mt_nonlinearity[0].name
            clientObject.diagram_around_x_symmetric = rotational_release_mt_nonlinearity[1][0]
            clientObject.diagram_around_x_is_sorted = True

            if rotational_release_mt_nonlinearity[1][0]:
                clientObject.diagram_around_x_start = rotational_release_mt_nonlinearity[1][1].name
                clientObject.diagram_around_x_end = rotational_release_mt_nonlinearity[1][1].name

            else:
                clientObject.diagram_around_x_start = rotational_release_mt_nonlinearity[1][1].name
                clientObject.diagram_around_x_end = rotational_release_mt_nonlinearity[1][2].name

            clientObject.diagram_around_x_table = model.clientModel.factory.create('ns0:member_hinge.diagram_around_x_table')

            for i,j in enumerate(rotational_release_mt_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:member_hinge_diagram_around_x_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:member_hinge_diagram_around_x_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.rotation = rotational_release_mt_nonlinearity[2][i][0]
                mlvlp.row.moment = rotational_release_mt_nonlinearity[2][i][1]

                clientObject.diagram_around_x_table.member_hinge_diagram_around_x_table.append(mlvlp)

        # Rotational Release My Nonlinearity
        if rotational_release_my_nonlinearity[0].name == "NONLINEARITY_TYPE_NONE" \
        or rotational_release_my_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE" \
        or rotational_release_my_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_POSITIVE" :
            clientObject.moment_release_my_nonlinearity = rotational_release_my_nonlinearity[0].name

        # Partial Activity
        elif rotational_release_my_nonlinearity[0].name == "NONLINEARITY_TYPE_PARTIAL_ACTIVITY":
            clientObject.moment_release_my_nonlinearity = rotational_release_my_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_around_y_negative_type = rotational_release_my_nonlinearity[1][0].name

            if rotational_release_my_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][1]

            elif rotational_release_my_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][1]
                clientObject.partial_activity_around_y_negative_rotation = rotational_release_my_nonlinearity[1][2]

            elif rotational_release_my_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT":
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][1]
                clientObject.partial_activity_around_y_negative_moment = rotational_release_my_nonlinearity[1][2]

            elif rotational_release_my_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT":
                clientObject.partial_activity_around_y_negative_slippage = rotational_release_my_nonlinearity[1][1]
                clientObject.partial_activity_around_y_negative_force = rotational_release_my_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_around_y_positive_type = rotational_release_my_nonlinearity[2][0].name

            if rotational_release_my_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][1]

            elif rotational_release_my_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][1]
                clientObject.partial_activity_around_y_positive_rotation = rotational_release_my_nonlinearity[2][2]

            elif rotational_release_my_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT":
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][1]
                clientObject.partial_activity_around_y_positive_moment = rotational_release_my_nonlinearity[2][2]

            elif rotational_release_my_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT":
                clientObject.partial_activity_around_y_positive_slippage = rotational_release_my_nonlinearity[2][1]
                clientObject.partial_activity_around_y_positive_moment = rotational_release_my_nonlinearity[2][2]

        elif rotational_release_my_nonlinearity[0].name == "NONLINEARITY_TYPE_DIAGRAM":
            clientObject.moment_release_my_nonlinearity = rotational_release_my_nonlinearity[0].name
            clientObject.diagram_around_y_symmetric = rotational_release_my_nonlinearity[1][0]
            clientObject.diagram_around_y_is_sorted = True

            if rotational_release_my_nonlinearity[1][0]:
                clientObject.diagram_around_y_start = rotational_release_my_nonlinearity[1][1].name
                clientObject.diagram_around_y_end = rotational_release_my_nonlinearity[1][1].name

            else:
                clientObject.diagram_around_y_start = rotational_release_my_nonlinearity[1][1].name
                clientObject.diagram_around_y_end = rotational_release_my_nonlinearity[1][2].name

            clientObject.diagram_around_y_table = model.clientModel.factory.create('ns0:member_hinge.diagram_around_y_table')

            for i,j in enumerate(rotational_release_my_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:member_hinge_diagram_around_y_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:member_hinge_diagram_around_y_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.rotation = rotational_release_my_nonlinearity[2][i][0]
                mlvlp.row.moment = rotational_release_my_nonlinearity[2][i][1]

                clientObject.diagram_around_y_table.member_hinge_diagram_around_y_table.append(mlvlp)

        # Rotational Release Mz Nonlinearity
        if rotational_release_mz_nonlinearity[0].name == "NONLINEARITY_TYPE_NONE" \
        or rotational_release_mz_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_NEGATIVE" \
        or rotational_release_mz_nonlinearity[0].name == "NONLINEARITY_TYPE_FAILURE_IF_POSITIVE" :
            clientObject.moment_release_mz_nonlinearity = rotational_release_mz_nonlinearity[0].name

        # Partial Activity
        elif rotational_release_mz_nonlinearity[0].name == "NONLINEARITY_TYPE_PARTIAL_ACTIVITY":
            clientObject.moment_release_mz_nonlinearity = rotational_release_mz_nonlinearity[0].name

            # Negative Zone
            clientObject.partial_activity_around_z_negative_type = rotational_release_mz_nonlinearity[1][0].name

            if rotational_release_mz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][1]

            elif rotational_release_mz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][1]
                clientObject.partial_activity_around_z_negative_rotation = rotational_release_mz_nonlinearity[1][2]

            elif rotational_release_mz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT":
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][1]
                clientObject.partial_activity_around_z_negative_moment = rotational_release_mz_nonlinearity[1][2]

            elif rotational_release_mz_nonlinearity[1][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT":
                clientObject.partial_activity_around_z_negative_slippage = rotational_release_mz_nonlinearity[1][1]
                clientObject.partial_activity_around_z_negative_force = rotational_release_mz_nonlinearity[1][2]

            # Positive Zone
            clientObject.partial_activity_around_z_positive_type = rotational_release_mz_nonlinearity[2][0].name

            if rotational_release_mz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_COMPLETE":
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][1]

            elif rotational_release_mz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FIXED":
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][1]
                clientObject.partial_activity_around_z_positive_rotation = rotational_release_mz_nonlinearity[2][2]

            elif rotational_release_mz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_FAILURE_FROM_MOMENT":
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][1]
                clientObject.partial_activity_around_z_positive_moment = rotational_release_mz_nonlinearity[2][2]

            elif rotational_release_mz_nonlinearity[2][0].name == "PARTIAL_ACTIVITY_TYPE_YIELDING_FROM_MOMENT":
                clientObject.partial_activity_around_z_positive_slippage = rotational_release_mz_nonlinearity[2][1]
                clientObject.partial_activity_around_z_positive_moment = rotational_release_mz_nonlinearity[2][2]

        elif rotational_release_mz_nonlinearity[0].name == "NONLINEARITY_TYPE_DIAGRAM":
            clientObject.moment_release_mz_nonlinearity = rotational_release_mz_nonlinearity[0].name
            clientObject.diagram_around_z_symmetric = rotational_release_mz_nonlinearity[1][0]
            clientObject.diagram_around_z_is_sorted = True

            if rotational_release_mz_nonlinearity[1][0]:
                clientObject.diagram_around_z_start = rotational_release_mz_nonlinearity[1][1].name
                clientObject.diagram_around_z_end = rotational_release_mz_nonlinearity[1][1].name

            else:
                clientObject.diagram_around_z_start = rotational_release_mz_nonlinearity[1][1].name
                clientObject.diagram_around_z_end = rotational_release_mz_nonlinearity[1][2].name

            clientObject.diagram_around_z_table = model.clientModel.factory.create('ns0:member_hinge.diagram_around_z_table')

            for i,j in enumerate(rotational_release_mz_nonlinearity[2]):
                mlvlp = model.clientModel.factory.create('ns0:member_hinge_diagram_around_z_table_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:member_hinge_diagram_around_z_table')
                clearAttributes(mlvlp.row)
                mlvlp.row.rotation = rotational_release_mz_nonlinearity[2][i][0]
                mlvlp.row.moment = rotational_release_mz_nonlinearity[2][i][1]

                clientObject.diagram_around_z_table.member_hinge_diagram_around_z_table.append(mlvlp)

        # Member HInge User defined name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Line to client model
        model.clientModel.service.set_member_hinge(clientObject)



from RSTAB.initModel import Model, clearAttributes, ConvertToDlString, deleteEmptyAttributes

class MemberResultIntermediatePoint():
    def __init__(self,
                 no: int = 1,
                 members: str = "",
                 point_count: int = 2,
                 uniform_distribution: bool = True,
                 distances = None,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Member Result Intermediate Point Tag
            members (str): Assigned Members
            point_count (int): Assigned Point Number
            uniform_distribution (bool): Uniform Distrubition Option
            distances (list of lists): Distances Table
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Member Result Intermediate Point
        clientObject = model.clientModel.factory.create('ns0:member_result_intermediate_point')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Result Intermediate Point No.
        clientObject.no = no

        # Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Point Count
        clientObject.uniform_distribution = uniform_distribution
        if uniform_distribution:
            clientObject.point_count = point_count

        else:
            clientObject.distances = model.clientModel.factory.create('ns0:member_result_intermediate_point.distances')

            for i,j in enumerate(distances):
                mlvlp = model.clientModel.factory.create('ns0:member_result_intermediate_point_distances_row')
                mlvlp.no = i+1
                mlvlp.row = model.clientModel.factory.create('ns0:member_result_intermediate_point_distances')
                mlvlp.row.value = distances[i][0]

                clientObject.distances.member_result_intermediate_point_distances.append(mlvlp)

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Result Intermediate Point to client model
        model.clientModel.service.set_member_result_intermediate_point(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import StaticAnalysisSettingsIterativeMethodForNonlinearAnalysis, StaticAnalysisType

class StaticAnalysisSettings():

    def __init__(self,
                 no: int = 1,
                 name: str = None,
                 analysis_type = StaticAnalysisType.GEOMETRICALLY_LINEAR,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Static Analysis Setting Tag
            name (str): Static Analysis Setting Name
            analysis_type (enum): Analysis Type Enumeration
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """
        # Client model | Surface
        clientObject = model.clientModel.factory.create('ns0:static_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Static Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Analysis Type
        clientObject.analysis_type = analysis_type.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Static Analysis Settings to client model
        model.clientModel.service.set_static_analysis_settings(clientObject)

    @staticmethod
    def GeometricallyLinear(
                  no: int = 1,
                  name: str = None,
                  load_modification: list = [False, 1, False],
                  bourdon_effect: bool = False,
                  mass_conversion: list = [False, 0, 0, 0],
                  comment: str = '',
                  params: dict = None,
                  model = Model):

        """
        Args:
            no (int): Static Analysis Setting Tag
            name (str, optional): Static Analysis Setting Name
            load_modification (list, optional): Load Modification Parameters
                load_modification = [loading_by_multiplier_factor, multiplier_factor, dividing_results]
            mass_conversion (list, optional): Mass Conversion Parameters
                mass_conversion = [mass_conversion_enabled, mass_conversion_factor_in_direction_x, mass_conversion_factor_in_direction_y, mass_conversion_factor_in_direction_z]
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model
        clientObject = model.clientModel.factory.create('ns0:static_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Static Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Static Analysis Type
        clientObject.analysis_type = StaticAnalysisType.GEOMETRICALLY_LINEAR.name

        # Bourdon Effect Displacement
        clientObject.displacements_due_to_bourdon_effect = bourdon_effect

        # Conversion of Mass into Load
        clientObject.mass_conversion_enabled = mass_conversion[0]
        if mass_conversion[0]:
            clientObject.mass_conversion_factor_in_direction_x = mass_conversion[1]
            clientObject.mass_conversion_factor_in_direction_y = mass_conversion[2]
            clientObject.mass_conversion_factor_in_direction_z = mass_conversion[3]

        if load_modification[0]:
            clientObject.modify_loading_by_multiplier_factor = True
            clientObject.loading_multiplier_factor = load_modification[1]
            clientObject.divide_results_by_loading_factor = load_modification[2]

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Static Analysis Settings to client model
        model.clientModel.service.set_static_analysis_settings(clientObject)

    @staticmethod
    def LargeDeformation(
                  no: int = 1,
                  name: str = None,
                  iterative_method = StaticAnalysisSettingsIterativeMethodForNonlinearAnalysis.NEWTON_RAPHSON,
                  precision_of_convergence_criteria_for_nonlinear_calculation: float = 0,
                  max_number_of_iterations: float = 100,
                  number_of_load_increments: float = 1,
                  load_modification: list = [False, 1, False],
                  bourdon_effect: bool = True,
                  mass_conversion: list = [False, 0, 0, 1],
                  comment: str = '',
                  params: dict = {'save_results_of_all_load_increments': False},
                  model = Model):

        """
        Args:
            no (int): Static Analysis Setting Tag
            name (str, optional):  Static Analysis Setting Name
            iterative_method (enum): Static Analysis Settings Iterative Method for Non-linear Analysis Enumeration
            precision_of_convergence_criteria_for_nonlinear_calculation (float): Precision of Convergence defaults to 0
            max_number_of_iterations (float): Maximum Number of Iterations
            number_of_load_increments (float): Number of Load Increments
            load_modification (list, optional): Load Modification Parameters
                load_modification = [loading_by_multiplier_factor, multiplier_factor, dividing_results]
            bourdon_effect (bool, optional): Bourdon Effect Boolean
            mass_conversion (list, optional): Mass Conversion Parameters
                mass_conversion = [mass_conversion_enabled, mass_conversion_factor_in_direction_x, mass_conversion_factor_in_direction_y, mass_conversion_factor_in_direction_z]
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model
        clientObject = model.clientModel.factory.create('ns0:static_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Static Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Static Analysis Type
        clientObject.analysis_type = StaticAnalysisType.LARGE_DEFORMATIONS.name

        # Bourdon Effect Displacement
        clientObject.displacements_due_to_bourdon_effect = bourdon_effect

        # Iterative Method
        clientObject.iterative_method_for_nonlinear_analysis = iterative_method.name

        clientObject.max_number_of_iterations = max_number_of_iterations
        clientObject.number_of_load_increments = number_of_load_increments

        # Conversion of Mass into Load
        clientObject.mass_conversion_enabled = mass_conversion[0]
        if mass_conversion[0]:
            clientObject.mass_conversion_factor_in_direction_x = mass_conversion[1]
            clientObject.mass_conversion_factor_in_direction_y = mass_conversion[2]
            clientObject.mass_conversion_factor_in_direction_z = mass_conversion[3]

        # Method for Equation System
        #clientObject.method_of_equation_system = method_of_equation_system.name

        # Modify Loading by Multiplier Factor
        if load_modification[0]:
            clientObject.modify_loading_by_multiplier_factor = True
            clientObject.loading_multiplier_factor = load_modification[1]
            clientObject.divide_results_by_loading_factor = load_modification[2]

        # Modify Standard precision and Tolerance Settings.
        if precision_of_convergence_criteria_for_nonlinear_calculation:
            clientObject.standard_precision_and_tolerance_settings_enabled = True
            clientObject.precision_of_convergence_criteria_for_nonlinear_calculation = precision_of_convergence_criteria_for_nonlinear_calculation

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Static Analysis Settings to client model
        model.clientModel.service.set_static_analysis_settings(clientObject)

    @staticmethod
    def SecondOrderPDelta(
                  no: int = 1,
                  name: str = None,
                  max_number_of_iterations: int = 100,
                  number_of_load_increments: int = 1,
                  load_modification: list = [False, 0, False],
                  precision_of_convergence_criteria_for_nonlinear_calculation: float = 0,
                  favorable_effect_due_to_tension_in_members : bool = False,
                  bourdon_effect: bool = True,
                  internal_forces_to_deformed_structure: list = [True, True, True, True],
                  mass_conversion: list = [False, 0, 0, 1],
                  comment: str = '',
                  params: dict = None,
                  model = Model):
        """
        Args:
            no (int): Static Analysis Setting Tag
            name (str):  Static Analysis Setting Name
            max_number_of_iterations (int): Maximum Number of Iterations
            number_of_load_increments (int): Number of Load Increments
            load_modification (list): Modify Loading by Multiplier Factor
                load_modification = [modify_loading_by_multiplier_factor, loading_multiplier_factor, divide_results_by_loading_factor]
            precision_of_convergence_criteria_for_nonlinear_calculation (float): Precision of Convergence defaults to 0
            favorable_effect_due_to_tension_in_members (bool): Favorable Effect due to Tension In Members Boolean
            bourdon_effect (bool): Bourdon Effect Boolean
            internal_forces_to_deformed_structure (list): Internal Forces to Deformed Structure List
                internal_forces_to_deformed_structure = [refer_internal_forces_to_deformed_structure, internal_forces_to_deformed_structure_for_moments, internal_forces_to_deformed_structure_for_normal_forces, internal_forces_to_deformed_structure_for_shear_forces]
            mass_conversion (list): Mass Conversion Parameters
                mass_conversion = [mass_conversion_enabled, mass_conversion_factor_in_direction_x, mass_conversion_factor_in_direction_y, mass_conversion_factor_in_direction_z]
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model
        clientObject = model.clientModel.factory.create('ns0:static_analysis_settings')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Static Analysis Settings No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Static Analysis Type
        clientObject.analysis_type = StaticAnalysisType.SECOND_ORDER_P_DELTA.name

        # Maximum Number of Iterations
        clientObject.max_number_of_iterations = max_number_of_iterations

        # Number of Load Increments
        clientObject.number_of_load_increments = number_of_load_increments

        # Modify Loading by Multiplier Factor
        if load_modification[0]:
            clientObject.modify_loading_by_multiplier_factor = True
            clientObject.loading_multiplier_factor = load_modification[1]
            clientObject.divide_results_by_loading_factor = load_modification[2]

        # Consider Favorable Effect due to Tension in Members
        clientObject.consider_favorable_effect_due_to_tension_in_members = favorable_effect_due_to_tension_in_members

        # Bourdon Effect Displacement
        clientObject.displacements_due_to_bourdon_effect = bourdon_effect

        # Conversion of Mass Into Load
        clientObject.mass_conversion_enabled = mass_conversion[0]
        if mass_conversion[0]:
            clientObject.mass_conversion_factor_in_direction_x = mass_conversion[1]
            clientObject.mass_conversion_factor_in_direction_y = mass_conversion[2]
            clientObject.mass_conversion_factor_in_direction_z = mass_conversion[3]

        # Iterative Method Settings
        if internal_forces_to_deformed_structure[0]:
            clientObject.refer_internal_forces_to_deformed_structure = True
            clientObject.refer_internal_forces_to_deformed_structure_for_moments = internal_forces_to_deformed_structure[1]
            clientObject.refer_internal_forces_to_deformed_structure_for_normal_forces = internal_forces_to_deformed_structure[2]
            clientObject.refer_internal_forces_to_deformed_structure_for_shear_forces = internal_forces_to_deformed_structure[3]

        # Modify Standard Precision and Tolerance Settings
        if precision_of_convergence_criteria_for_nonlinear_calculation:
            clientObject.standard_precision_and_tolerance_settings_enabled = True
            clientObject.precision_of_convergence_criteria_for_nonlinear_calculation = precision_of_convergence_criteria_for_nonlinear_calculation

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Static Analysis Settings to client model
        model.clientModel.service.set_static_analysis_settings(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import CoordinateSystemType, CoordinateSystemRotationAnglesSequence

class CoordinateSystem():
    def __init__(self,
                 no: int = 1,
                 origin_coordinate_x: float = 0.0,
                 origin_coordinate_y: float = 0.0,
                 origin_coordinate_z: float = 0.0,
                 name: str = 'Coord1',
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
         Args:
            no (int): Coordinate System Tag
            origin_coordinate_x (float): X-Coordinate
            origin_coordinate_y (float): Y-Coordinate
            origin_coordinate_z (float): Z-Coordinate
            name (str, optional): User Defined Coordinate System Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Coordinate System
        clientObject = model.clientModel.factory.create('ns0:coordinate_system')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Coordinate System No.
        clientObject.no = no

        # Coordinate System Type
        clientObject.type = CoordinateSystemType.TYPE_OFFSET_XYZ.name

        # Coordinates
        clientObject.origin_coordinate_x = origin_coordinate_x
        clientObject.origin_coordinate_y = origin_coordinate_y
        clientObject.origin_coordinate_z = origin_coordinate_z

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Coordinate System to client model
        model.clientModel.service.set_coordinate_system(clientObject)

    @staticmethod
    def OffsetXYZ(no: int = 1,
                  origin_coordinate_x: float = 0.0,
                  origin_coordinate_y: float = 0.0,
                  origin_coordinate_z: float = 0.0,
                  name: str = '',
                  comment: str = '',
                  params: dict = None,
                  model = Model):

        '''
         Args:
            no (int): Coordinate System Tag
            origin_coordinate_x (float): X-Coordinate
            origin_coordinate_y (float): Y-Coordinate
            origin_coordinate_z (float): Z-Coordinate
            name (str, optional): User Defined Coordinate System Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Coordinate System
        clientObject = model.clientModel.factory.create('ns0:coordinate_system')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Coordinate System No.
        clientObject.no = no

        # Coordinate System Type
        clientObject.type = CoordinateSystemType.TYPE_OFFSET_XYZ.name

        # Coordinates
        clientObject.origin_coordinate_x = origin_coordinate_x
        clientObject.origin_coordinate_y = origin_coordinate_y
        clientObject.origin_coordinate_z = origin_coordinate_z

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Coordinate System to client model
        model.clientModel.service.set_coordinate_system(clientObject)

    @staticmethod
    def ThreePoints(no: int = 1,
                    origin_coordinate_x: float = 1.0,
                    origin_coordinate_y: float = 0.0,
                    origin_coordinate_z: float = 0.0,
                    u_axis_point_coordinate_x: float = 0.0,
                    u_axis_point_coordinate_y: float = 1.0,
                    u_axis_point_coordinate_z: float = 0.0,
                    uw_plane_point_coordinate_x: float = 0.0,
                    uw_plane_point_coordinate_y: float = 0.0,
                    uw_plane_point_coordinate_z: float = 1.0,
                    name: str = '',
                    comment: str = '',
                    params: dict = None,
                    model = Model):

        '''
         Args:
            no (int): Coordinate System Tag
            origin_coordinate_x (float): Origin Point X-Coordinate
            origin_coordinate_y (float): Origin Point Y-Coordinate
            origin_coordinate_z (float): Origin Point Z-Coordinate
            u_axis_point_coordinate_x (float): Point on +U-Axis - 1st point X-Coordinate
            u_axis_point_coordinate_y (float): Point on +U-Axis - 1st point Y-Coordinate
            u_axis_point_coordinate_z (float): Point on +U-Axis - 1st point Z-Coordinate
            uw_plane_point_coordinate_x (float): Point in +UW-Plane - 2nd Point X-Coordinate
            uw_plane_point_coordinate_y (float): Point in +UW-Plane - 2nd Point Y-Coordinate
            uw_plane_point_coordinate_z (float): Point in +UW-Plane - 2nd Point Z-Coordinate
            name (str, optional): User Defined Coordinate System Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Coordinate System
        clientObject = model.clientModel.factory.create('ns0:coordinate_system')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Coordinate System No.
        clientObject.no = no

        # Coordinate System Type
        clientObject.type = CoordinateSystemType.TYPE_3_POINTS.name

        # Coordinates
        clientObject.origin_coordinate_x = origin_coordinate_x
        clientObject.origin_coordinate_y = origin_coordinate_y
        clientObject.origin_coordinate_z = origin_coordinate_z

        clientObject.u_axis_point_coordinate_x = u_axis_point_coordinate_x
        clientObject.u_axis_point_coordinate_y = u_axis_point_coordinate_y
        clientObject.u_axis_point_coordinate_z = u_axis_point_coordinate_z

        clientObject.uw_plane_point_coordinate_x = uw_plane_point_coordinate_x
        clientObject.uw_plane_point_coordinate_y = uw_plane_point_coordinate_y
        clientObject.uw_plane_point_coordinate_z = uw_plane_point_coordinate_z

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Coordinate System to client model
        model.clientModel.service.set_coordinate_system(clientObject)

    @staticmethod
    def TwoPointsAndAngle(no: int = 1,
                          origin_coordinate_x: float = 1.0,
                          origin_coordinate_y: float = 0.0,
                          origin_coordinate_z: float = 0.0,
                          u_axis_point_coordinate_x: float = 0.0,
                          u_axis_point_coordinate_y: float = 0.0,
                          u_axis_point_coordinate_z: float = 0.0,
                          uw_plane_angle: float = 0.0,
                          name: str = '',
                          comment: str = '',
                          params: dict = None,
                          model = Model):

        '''
         Args:
            no (int): Coordinate System Tag
            origin_coordinate_x (float): Origin Point X-Coordinate
            origin_coordinate_y (float): Origin Point Y-Coordinate
            origin_coordinate_z (float): Origin Point Z-Coordinate
            u_axis_point_coordinate_x (float): Point on +U-Axis - 1st point X-Coordinate
            u_axis_point_coordinate_y (float): Point on +U-Axis - 1st point Y-Coordinate
            u_axis_point_coordinate_z (float): Point on +U-Axis - 1st point Z-Coordinate
            uw_plane_angle (float): Rotation About U-Axis
            name (str, optional): User Defined Coordinate System Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Coordinate System
        clientObject = model.clientModel.factory.create('ns0:coordinate_system')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Coordinate System No.
        clientObject.no = no

        # Coordinate System Type
        clientObject.type = CoordinateSystemType.TYPE_2_POINTS_AND_ANGLE.name

        # Coordinates
        clientObject.origin_coordinate_x = origin_coordinate_x
        clientObject.origin_coordinate_y = origin_coordinate_y
        clientObject.origin_coordinate_z = origin_coordinate_z

        clientObject.u_axis_point_coordinate_x = u_axis_point_coordinate_x
        clientObject.u_axis_point_coordinate_y = u_axis_point_coordinate_y
        clientObject.u_axis_point_coordinate_z = u_axis_point_coordinate_z

        clientObject.uw_plane_angle = uw_plane_angle

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Coordinate System to client model
        model.clientModel.service.set_coordinate_system(clientObject)

    @staticmethod
    def PointAndThreeAngles(no: int = 1,
                            origin_coordinate_x: float = 1.0,
                            origin_coordinate_y: float = 2.0,
                            origin_coordinate_z: float = 3.0,
                            rotation_angles_sequence: float = CoordinateSystemRotationAnglesSequence.SEQUENCE_XYZ,
                            rotation_angle_1: float = 0.0,
                            rotation_angle_2: float = 0.0,
                            rotation_angle_3: float = 0.0,
                            name: str = '',
                            comment: str = '',
                            params: dict = None,
                            model = Model):

        '''
         Args:
            no (int): Coordinate System Tag
            origin_coordinate_x (float): Origin Point X-Coordinate
            origin_coordinate_y (float): Origin Point Y-Coordinate
            origin_coordinate_z (float): Origin Point Z-Coordinate
            rotation_angles_sequence (enum): Coordinate System Rotation Angles Sequence Enumeration
            rotation_angle_1 (float): Rotation about X Axes
            rotation_angle_2 (float): Rotation about Y Axes
            rotation_angle_3 (float): Rotation about Z Axes
            name (str, optional): User Defined Coordinate System Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''
        # Client model | Coordinate System
        clientObject = model.clientModel.factory.create('ns0:coordinate_system')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Coordinate System No.
        clientObject.no = no

        # Coordinate System Type
        clientObject.type = CoordinateSystemType.TYPE_POINT_AND_3_ANGLES.name

        # Coordinates
        clientObject.origin_coordinate_x = origin_coordinate_x
        clientObject.origin_coordinate_y = origin_coordinate_y
        clientObject.origin_coordinate_z = origin_coordinate_z

        # Rotation Angles
        clientObject.rotation_angles_sequence = rotation_angles_sequence.name

        clientObject.rotation_angle_1 = rotation_angle_1
        clientObject.rotation_angle_2 = rotation_angle_2
        clientObject.rotation_angle_3 = rotation_angle_3

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Coordinate System to client model
        model.clientModel.service.set_coordinate_system(clientObject)

    @staticmethod
    def GetCoordinateSystem(object_index: int = 1, model = Model):

        '''
        Args:
            obejct_index (int): Coordinate System Index
            model (RSTAB Class, optional): Model to be edited
        '''

        # Get Coordinate System from client model
        return model.clientModel.service.get_coordinate_system(object_index)



import sys
import enum
from RSTAB.initModel import Model
from RSTAB.enums import PlausibilityCheckType

class PlausibilityCheck():

    def __init__(self,
                 plausibility_check_type:enum = PlausibilityCheckType.PLAUSIBILITY_CHECK,
                 skip_warnings:bool = False,
                 model = Model):

        response = model.clientModel.service.plausibility_check(plausibility_check_type.name, skip_warnings)

        if response["errors_and_warnings"] and response["errors_and_warnings"][0] and not skip_warnings:
            errormessage = ''
            for item in response["errors_and_warnings"][0]:
                errors_and_warnings = ''
                if item.message_type == "ERROR":
                    errors_and_warnings = "".join([item.message_type, '!\nObject: ', item.input_field,", ", item.object,", current value: ", item.current_value,", message: ", item.message])
                elif item.message_type == "WARNING":
                    errors_and_warnings =  "".join([item.message_type, '!\n', item.message])
                errormessage += '\n'+errors_and_warnings.replace('<br>','\n')

            sys.exit(errormessage)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString
from RSTAB.enums import DesignSupportOrientationZType, DesignSupportOrientationYType

class DesignSupport():

    def __init__(self,
                 no: int = 1,
                 members: str = '1',
                 member_sets: str = None,
                 nodes: str = '1 2',
                 support_in_z: list = [True, True, DesignSupportOrientationZType.DESIGN_SUPPORT_ORIENTATION_ZAXIS_POSITIVE, 0.1],
                 support_in_y: list = None,
                 name: str = None,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Design Support Type General

        Args:
            no (int): Design Support Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            nodes (str): Assigned Nodes
            support_in_z (list/bool): List of Parameters for Support in Z Axis
            support_in_y (list/bool): List of Parameters for Support in Y Axis
                for deactive support in z/y:
                    support_in_z/y (bool): Deactivate support in Z/Y (= False/None)
                for active support in z/y:
                    support_in_z/y[0] (bool): Activate Support in Z/Y (= True)
                    support_in_z/y[1] (bool): Consider Support in Deflection Design Z/Y
                    support_in_z/y[2] (enum): Design Support Orientation Z/Y Type Enumeration
                    support_in_z/y[3] (float): Support Width Z/Y (in meter)
                    for manual support depth:
                        support_in_z/y[4] (float): Support Depth Z/Y (in meter)
            name (str, option): User Defined Design Support Name
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Design Support
        clientObject = model.clientModel.factory.create('ns0:design_support')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Design Support No.
        clientObject.no = no

        # Design Support Type
        clientObject.type = 'DESIGN_SUPPORT_TYPE_GENERAL'

        # Assigned Members (e.g. '5 6 7 12')
        if members:
            clientObject.assigned_to_members = ConvertToDlString(members)

        # Assigned Member Sets (e.g. '5 6 7 12')
        if member_sets:
            clientObject.assigned_to_member_sets = ConvertToDlString(member_sets)

        # Assigned Nodes (e.g. '5 6 7 12')
        clientObject.assigned_to_nodes = ConvertToDlString(nodes)

        # Support in Z
        if isinstance(support_in_z, list) and support_in_z[0] == True:
            clientObject.activate_in_z = support_in_z[0]
            clientObject.consider_in_deflection_design_z = support_in_z[1]
            clientObject.design_support_orientation_z = support_in_z[2].name
            clientObject.support_width_z = support_in_z[3]

            if len(support_in_z) == 5:
                clientObject.support_depth_by_section_width_of_member_z_enabled = False
                clientObject.support_depth_z = support_in_z[4]

            elif len(support_in_z) == 4:
                clientObject.support_depth_by_section_width_of_member_z_enabled = True

            else:
                raise ValueError('WARNING!: The suport in Z parameter needs to be length of 4 or 5. Kindly check list inputs for completeness and correctness.')

        elif support_in_z == None or support_in_z == False:
            clientObject.activate_in_z = False

        else:
            raise TypeError("WARNING! First parameter of list must be type bool. Kindly check list inputs completeness and correctness.")

        # Support in Y
        if isinstance(support_in_y, list) and support_in_y[0] == True:
            clientObject.activate_in_y = support_in_y[0]
            clientObject.consider_in_deflection_design_y = support_in_y[1]
            clientObject.design_support_orientation_y = support_in_y[2].name
            clientObject.support_width_y = support_in_y[3]

            if len(support_in_y) == 5:
                clientObject.support_depth_by_section_width_of_member_y_enabled = False
                clientObject.support_depth_y = support_in_y[4]

            elif len(support_in_y) == 4:
                clientObject.support_depth_by_section_width_of_member_y_enabled = True

            else:
                raise ValueError('WARNING!: The suport in Y parameter needs to be length of 4 or 5. Kindly check list inputs for completeness and correctness.')

        elif support_in_y == None or support_in_y == False:
            clientObject.activate_in_y = False

        else:
            raise TypeError("WARNING! First parameter of list must be type bool. Kindly check list inputs completeness and correctness.")

        # User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Design Support to client model
        model.clientModel.service.set_design_support(clientObject)

    @staticmethod
    def Concrete(no: int = 1,
                 members: str = '1',
                 member_sets: str = None,
                 nodes: str = '1 2',
                 support_in_z: list = [True, True, True, True, 1.0, 0.1],
                 support_in_y: bool = True,
                 name: str = None,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        '''
        Design Support Type Concrete

        Args:
            no (int): Design Support Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            nodes (str): Assigned Nodes
            support_in_z (list): List of Parameters for Support in Z Axis
                for deactive support in z:
                    support_in_z (bool): Deactivate support in Z (= False/None)
                for active support in z:
                    support_in_z[0] (bool): Activate Support in Z (= True)
                    support_in_z[1] (bool): Consider Support in Deflection Design Z
                    support_in_z[2] (bool): Enable/Disable Direct Support Z
                    support_in_z[3] (bool): Enable/Disable Concrete Monolithic Connection Z
                    for inner_support_z_enabled == True:
                        support_in_z[4] (float): Concrete Ratio of Moment Redistribution Z (It must be in between 0.65 and 1.0)
                    for inner_support_z_enabled == False:
                        support_in_z[4] (bool): Deactivate Inner Support in Z (= False)
                    support_in_z[5] (float): Support Width Z (in meter)
                    for manual support depth:
                        support_in_z[6] (float): Support Depth Z (in meter)
            support_in_y (bool): Consider Support in Y Axis for Deflection Design (= True/False)
            name (str, option): User Defined Design Support Name
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Design Support
        clientObject = model.clientModel.factory.create('ns0:design_support')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Design Support No.
        clientObject.no = no

        # Design Support Type
        clientObject.type = 'DESIGN_SUPPORT_TYPE_CONCRETE'

        # Assigned Members (e.g. '5 6 7 12')
        if members:
            clientObject.assigned_to_members = ConvertToDlString(members)

        # Assigned Member Sets (e.g. '5 6 7 12')
        if member_sets:
            clientObject.assigned_to_member_sets = ConvertToDlString(member_sets)

        # Assigned Nodes (e.g. '5 6 7 12')
        clientObject.assigned_to_nodes = ConvertToDlString(nodes)

        # Support in Z
        if isinstance(support_in_z, list) and support_in_z[0] == True:
            clientObject.activate_in_z = support_in_z[0]
            clientObject.consider_in_deflection_design_z = support_in_z[1]
            clientObject.direct_support_z_enabled = support_in_z[2]
            clientObject.concrete_monolithic_connection_z_enabled = support_in_z[3]

            if isinstance(support_in_z[4], bool) and support_in_z[4] == False:
                clientObject.inner_support_z_enabled = False

            elif isinstance(support_in_z[4], (float, int)) and (0.65 <= support_in_z[4] <= 1):
                clientObject.inner_support_z_enabled = True
                clientObject.concrete_ratio_of_moment_redistribution_z = support_in_z[4]

            else:
                raise ValueError('WARNING!: The concrete ratio of moment redistribution in Z parameter must be between 0.65 and 1.0. Kindly check list inputs for completeness and correctness.')

            clientObject.support_width_z = support_in_z[5]

            if len(support_in_z) == 7:
                clientObject.support_depth_by_section_width_of_member_z_enabled = False
                clientObject.support_depth_z = support_in_z[6]

            elif len(support_in_z) == 6:
                clientObject.support_depth_by_section_width_of_member_z_enabled = True

            else:
                raise ValueError('WARNING!: The suport in Z parameter needs to be length of 6 or 7. Kindly check list inputs for completeness and correctness.')

        elif support_in_z == None or support_in_z == False:
            clientObject.activate_in_z = False

        else:
            raise TypeError("WARNING! First parameter of list must be type bool. Kindly check list inputs completeness and correctness.")

        # Support in Y
        if isinstance(support_in_y, bool) and support_in_y == True:
            clientObject.activate_in_y = True
            clientObject.consider_in_deflection_design_y = True

        elif support_in_y == None or support_in_y == False:
            clientObject.activate_in_y = False
            clientObject.consider_in_deflection_design_y = False

        else:
            raise TypeError("WARNING! Parameter must be type bool. Kindly check input completeness and correctness.")

        # User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Design Support to client model
        model.clientModel.service.set_design_support(clientObject)

    @staticmethod
    def Steel(no: int = 1,
              members: str = '1',
              member_sets: str = None,
              nodes: str = '1 2',
              support_in_z: list = [True, True, DesignSupportOrientationZType.DESIGN_SUPPORT_ORIENTATION_ZAXIS_POSITIVE, 0.01, 0.2, 0.25],
              support_in_y: list = None,
              name: str = None,
              comment: str = '',
              params: dict = None,
              model = Model):

        '''
        Design Support Type Steel

        Args:
            no (int): Design Support Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            nodes (str): Assigned Nodes
            support_in_z (list/bool): List of Parameters for Support in Z Axis
                for deactive support in z:
                    support_in_z (bool): Deactivate support in Z (= False/None)
                for active support in z:
                    support_in_z[0] (bool): Activate Support in Z (= True)
                    support_in_z[1] (bool): Consider Support in Deflection Design Z
                    support_in_z[2] (enum): Design Support Orientation Z Type Enumeration
                    for end_support_z_enabled == True:
                        support_in_z[3] (float): Overhanging Length for Support Z (in meter)
                    for end_support_z_enabled == False:
                        support_in_z[3] (bool): Deactivate End Support in Z (= False)
                    support_in_z[4] (float): Support Width Z (in meter)
                    for manual support depth:
                        support_in_z[5] (float): Support Depth Z (in meter)
            support_in_y (bool): Consider Support in Y Axis for Deflection Design (= True/False)
            name (str, option): User Defined Design Support Name
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Design Support
        clientObject = model.clientModel.factory.create('ns0:design_support')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Design Support No.
        clientObject.no = no

        # Design Support Type
        clientObject.type = 'DESIGN_SUPPORT_TYPE_STEEL'

        # Assigned Members (e.g. '5 6 7 12')
        if members:
            clientObject.assigned_to_members = ConvertToDlString(members)

        # Assigned Member Sets (e.g. '5 6 7 12')
        if member_sets:
            clientObject.assigned_to_member_sets = ConvertToDlString(member_sets)

        # Assigned Nodes (e.g. '5 6 7 12')
        clientObject.assigned_to_nodes = ConvertToDlString(nodes)

        # Support in Z
        if isinstance(support_in_z, list) and support_in_z[0] == True:
            clientObject.activate_in_z = support_in_z[0]
            clientObject.consider_in_deflection_design_z = support_in_z[1]
            clientObject.design_support_orientation_z = support_in_z[2].name

            if isinstance(support_in_z[3], bool) and support_in_z[3] == False:
                clientObject.end_support_z_enabled = False

            elif isinstance(support_in_z[3], (float, int)):
                clientObject.end_support_z_enabled = True
                clientObject.overhang_length_z = support_in_z[3]

            else:
                raise ValueError('WARNING!: Invalid list of parameters. Kindly check list inputs for completeness and correctness.')

            clientObject.support_width_z = support_in_z[4]

            if len(support_in_z) == 6:
                clientObject.support_depth_by_section_width_of_member_z_enabled = False
                clientObject.support_depth_z = support_in_z[5]

            elif len(support_in_z) == 5:
                clientObject.support_depth_by_section_width_of_member_z_enabled = True

            else:
                raise ValueError('WARNING!: The suport in Z parameter needs to be length of 5 or 6. Kindly check list inputs for completeness and correctness.')

        elif support_in_z == None or support_in_z == False:
            clientObject.activate_in_z = False

        else:
            raise TypeError("WARNING! First parameter of list must be type bool. Kindly check list inputs completeness and correctness.")

        # Support in Y
        if isinstance(support_in_y, bool) and support_in_y == True:
            clientObject.consider_in_deflection_design_y = True

        elif support_in_y == None or support_in_y == False:
            clientObject.consider_in_deflection_design_y = False

        else:
            raise TypeError("WARNING! Parameter must be type bool. Kindly check input completeness and correctness.")

        # User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Design Support to client model
        model.clientModel.service.set_design_support(clientObject)

    @staticmethod
    def Timber(no: int = 1,
               members: str = '1',
               member_sets: str = None,
               nodes: str = '1 2',
               support_in_z: list = [True, True, True, True, DesignSupportOrientationZType.DESIGN_SUPPORT_ORIENTATION_ZAXIS_POSITIVE, True, 1.25, 0.2, 0.25],
               support_in_y: list = [True, False, True, True, DesignSupportOrientationYType.DESIGN_SUPPORT_ORIENTATION_YAXIS_BOTH],
               name: str = None,
               comment: str = '',
               params: dict = None,
               model = Model):

        '''
        Design Support Type Timber

        Args:
            no (int): Design Support Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            nodes (str): Assigned Nodes
            support_in_z (list/bool): List of Parameters for Support in Z Axis
            support_in_y (list/bool): List of Parameters for Support in Y Axis
                for deactive support in z/y:
                    support_in_z/y (bool): Deactivate support in Z/Y (= False/None)
                for active support in z/y:
                    support_in_z/y[0] (bool): Activate Support in Z/Y (= True)
                    support_in_z/y[1] (bool): Enable/Disable Direct Support in Z/Y
                    support_in_z/y[2] (bool): Enable/Disable Inner Support in Z/Y
                    support_in_z/y[3] (bool): Consider Support in Deflection Design Z/Y
                    support_in_z/y[4] (enum): Design Support Orientation Z/Y Type Enumeration
                --> if direct support in z/y is enabled: (support_in_z/y[1] == True)
                        support_in_z/y[5] (bool): Consider Support in Fire Design Z/Y
                        support_in_z/y[6] (float): Timber Factor of Compression Z/Y (It must be in between 1.0 and 2.2)
                        support_in_z/y[7] (float): Support Width Z/Y (in meter)
                        for manual support depth:
                            support_in_z/y[8] (float): Support Depth Z/Y (in meter)
            name (str, option): User Defined Design Support Name
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        '''

        # Client model | Design Support
        clientObject = model.clientModel.factory.create('ns0:design_support')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Design Support No.
        clientObject.no = no

        # Design Support Type
        clientObject.type = 'DESIGN_SUPPORT_TYPE_TIMBER'

        # Assigned Members (e.g. '5 6 7 12')
        if members:
            clientObject.assigned_to_members = ConvertToDlString(members)

        # Assigned Member Sets (e.g. '5 6 7 12')
        if member_sets:
            clientObject.assigned_to_member_sets = ConvertToDlString(member_sets)

        # Assigned Nodes (e.g. '5 6 7 12')
        clientObject.assigned_to_nodes = ConvertToDlString(nodes)

        # Support in Z
        if isinstance(support_in_z, list) and support_in_z[0] == True:
            clientObject.activate_in_z = support_in_z[0]
            clientObject.direct_support_z_enabled = support_in_z[1]
            clientObject.inner_support_z_enabled = support_in_z[2]
            clientObject.consider_in_deflection_design_z = support_in_z[3]
            clientObject.design_support_orientation_z = support_in_z[4].name

            if support_in_z[1] == True:
                clientObject.consider_in_fire_design_z = support_in_z[5]

                if (1.0 <= support_in_z[6] <= 2.20):
                    clientObject.timber_factor_of_compression_z = support_in_z[6]

                else:
                    raise ValueError('WARNING!: The timber factor of compression in Z parameter must be between 1.0 and 2.2. Kindly check list inputs for completeness and correctness.')

                clientObject.support_width_z = support_in_z[7]

                if len(support_in_z) == 9:
                    clientObject.support_depth_by_section_width_of_member_z_enabled = False
                    clientObject.support_depth_z = support_in_z[8]

                elif len(support_in_z) == 8:
                    clientObject.support_depth_by_section_width_of_member_z_enabled = True

                else:
                    raise ValueError('WARNING!: The suport in Z parameter needs to be length of 8 or 9. Kindly check list inputs for completeness and correctness.')

        elif support_in_z == None or support_in_z == False:
            clientObject.activate_in_z = False

        else:
            raise TypeError("WARNING! First parameter of support in Z list must be type bool. Kindly check list inputs completeness and correctness.")

        # Support in Y
        if isinstance(support_in_y, list) and support_in_y[0] == True:
            clientObject.activate_in_y = support_in_y[0]
            clientObject.direct_support_y_enabled = support_in_y[1]
            clientObject.inner_support_y_enabled = support_in_y[2]
            clientObject.consider_in_deflection_design_y = support_in_y[3]
            clientObject.design_support_orientation_y = support_in_y[4].name

            if support_in_y[1] == True:
                clientObject.consider_in_fire_design_y = support_in_y[5]

                if (1.0 <= support_in_y[6] <= 2.20):
                    clientObject.timber_factor_of_compression_y = support_in_y[6]

                else:
                    raise ValueError('WARNING!: The timber factor of compression in Y parameter must be between 1.0 and 2.2. Kindly check list inputs for completeness and correctness.')

                clientObject.support_width_y = support_in_y[7]

                if len(support_in_y) == 9:
                    clientObject.support_depth_by_section_width_of_member_y_enabled = False
                    clientObject.support_depth_y = support_in_y[8]

                elif len(support_in_y) == 8:
                    clientObject.support_depth_by_section_width_of_member_y_enabled = True

                else:
                    raise ValueError('WARNING!: The suport in Y parameter needs to be length of 8 or 9. Kindly check list inputs for completeness and correctness.')

        elif support_in_y == None or support_in_y == False:
            clientObject.activate_in_y = False

        else:
            raise TypeError("WARNING! First parameter of support in Y list must be type bool. Kindly check list inputs completeness and correctness.")

        # User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Design Support to client model
        model.clientModel.service.set_design_support(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertStrToListOfInt
from RSTAB.enums import WeldComponentType, MultipleOffsetDefinitionType, WeldingMethod

transverseWeldComponent = {
    'weld_type' : WeldComponentType.WELD_COMPONENT_TYPE_BUTT,
    'position' : 0.3,
    'multiple' : True,
    'multiple_number' : 2,
    'multiple_offset_definition_type' : MultipleOffsetDefinitionType.OFFSET_DEFINITION_TYPE_ABSOLUTE,
    'multiple_offset' : 2.0,
    'size' : 0.005,
    'method_type' : WeldingMethod.WELDING_METHOD_MIG,
    'number_of_heat_paths' : 2,
    'welding_temperature' : 2000.0
}

class AluminumMemberTransverseWeld():

    def __init__(self,
                 no: int = 1,
                 name: str = '',
                 members: str = '1 2',
                 member_sets: str = '',
                 component: list = [transverseWeldComponent],
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Line Welded Joint Tag
            name (str): Name
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            component (struct): Weld Components
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Aluminum Member Transverse Weld
        clientObject = model.clientModel.factory.create('ns0:aluminum_member_transverse_weld')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Weld No.
        clientObject.no = no

        # Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Assigned to members
        clientObject.members = ConvertStrToListOfInt(members)

        # Assigned to member sets
        clientObject.member_sets = ConvertStrToListOfInt(member_sets)

        # Weld Components
        clientObject.components = model.clientModel.factory.create('ns0:array_of_aluminum_member_transverse_weld_components')

        count = 0
        for comp in component:
            clientWeld = model.clientModel.factory.create('ns0:aluminum_member_transverse_weld_components_row')
            clientWeld.no = count+1
            count = count+1
            clientWeld.row = model.clientModel.factory.create('ns0:aluminum_member_transverse_weld_components')
            clearAttributes(clientWeld.row)
            clientWeld.row.weld_type = comp['weld_type'].name
            clientWeld.row.position = comp['position']
            clientWeld.row.multiple = comp['multiple']
            clientWeld.row.multiple_number = comp['multiple_number']
            clientWeld.row.multiple_offset_definition_type = comp['multiple_offset_definition_type'].name
            clientWeld.row.multiple_offset = comp['multiple_offset']
            clientWeld.row.size = comp['size']
            clientWeld.row.method_type = comp['method_type'].name
            clientWeld.row.number_of_heat_paths = comp['number_of_heat_paths']
            clientWeld.row.welding_temperature = comp['welding_temperature']

            clientObject.components.aluminum_member_transverse_weld_components.append(clientWeld)

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Aluminum Member Transverse Weld to client model
        model.clientModel.service.set_aluminum_member_transverse_weld(clientObject)



from RSTAB.enums import MemberStiffnessModificationType
from RSTAB.initModel import Model, clearAttributes, ConvertToDlString, deleteEmptyAttributes

class MemberStiffnessModification():
    def __init__(self,
                 no: int = 1,
                 assigned_structure_modification: str = "",
                 modification_type = MemberStiffnessModificationType.TYPE_TOTAL_STIFFNESSES_FACTORS,
                 parameters = [1.0],
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Modification Tag
            assigned_structure_modification (str): Assigned Structure Modification
            modification_type (enum): Member Stiffness Modification Type Enumeration Item
            parameters (list): Parameters List
                for modification_type == "TYPE_TOTAL_STIFFNESSES_FACTORS":
                    parameters = [total_stiffness_factor]
                for modification_type == "TYPE_PARTIAL_STIFFNESSES_FACTORS":
                    parameters = [factor_of_axial_stiffness, factor_of_bending_y_stiffness, factor_of_bending_z_stiffness, partial_stiffness_factor_of_shear_y_stiffness,
                                  partial_stiffness_factor_of_shear_z_stiffness, partial_stiffness_factor_of_torsion_stiffness, partial_stiffness_factor_of_weight]
                for modification_type == "TYPE_CONCRETE_STRUCTURES_ACI":
                    parameters = [concrete_structure_component_type, factor_of_axial_stiffness, factor_of_bending_y_stiffness, factor_of_bending_z_stiffness]
                for modification_type == "TYPE_CONCRETE_STRUCTURES_CSA":
                    parameters = [concrete_structure_component_type, factor_of_axial_stiffness, factor_of_bending_y_stiffness, factor_of_bending_z_stiffness]
                for modification_type = "TYPE_STEEL_STRUCTURES":
                    parameters = [steel_structure_determine_tau_b, steel_structure_design_method]
                for modification_type = "TYPE_STEEL_STRUCTURES_CSA":
                    parameters = [steel_structure_csa_determine_tau_b, factor_of_axial_stiffness, factor_of_bending_z_stiffness, steel_structure_csa_stiffness_factor_of_shear_y_stiffness,
                                  steel_structure_csa_stiffness_factor_of_shear_z_stiffness, steel_structure_csa_stiffness_factor_of_torsion_stiffness]
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Member Stiffness Modification
        clientObject = model.clientModel.factory.create('ns0:member_stiffness_modification')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Member Stiffness Modification No.
        clientObject.no = no

        # Structure Modification Assignment
        clientObject.assigned_to_structure_modification = ConvertToDlString(assigned_structure_modification)

        # Stiffness Modification Type
        clientObject.type = modification_type.name

        if modification_type.name == "TYPE_TOTAL_STIFFNESSES_FACTORS":
            clientObject.factor_of_axial_stiffness = parameters[0]
            clientObject.factor_of_bending_y_stiffness = parameters[0]
            clientObject.factor_of_bending_z_stiffness = parameters[0]
            clientObject.total_stiffness_factor_of_total_stiffness = parameters[0]

        elif modification_type.name == "TYPE_PARTIAL_STIFFNESSES_FACTORS":
            clientObject.factor_of_axial_stiffness = parameters [0]
            clientObject.factor_of_bending_y_stiffness = parameters [2]
            clientObject.factor_of_bending_z_stiffness = parameters [3]
            clientObject.partial_stiffness_factor_of_shear_y_stiffness = parameters [4]
            clientObject.partial_stiffness_factor_of_shear_z_stiffness = parameters [5]
            clientObject.partial_stiffness_factor_of_torsion_stiffness = parameters [6]
            clientObject.partial_stiffness_factor_of_weight = parameters [7]

        elif modification_type.name == "TYPE_CONCRETE_STRUCTURES_ACI":
            clientObject.concrete_structure_component_type = parameters[0].name
            clientObject.factor_of_axial_stiffness = parameters[1]
            clientObject.factor_of_bending_y_stiffness = parameters[2]
            clientObject.factor_of_bending_z_stiffness = parameters[3]

        elif modification_type.name == "TYPE_CONCRETE_STRUCTURES_CSA":
            clientObject.concrete_structure_component_type = parameters[0].name
            clientObject.factor_of_axial_stiffness = parameters[1]
            clientObject.factor_of_bending_y_stiffness = parameters[2]
            clientObject.factor_of_bending_z_stiffness = parameters[3]

        elif modification_type.name == "TYPE_STEEL_STRUCTURES":
            clientObject.steel_structure_determine_tau_b = parameters[0]

            if parameters[0] == "ITERATIVE":
                clientObject.steel_structure_design_method = parameters[1]

        elif modification_type.name == "TYPE_STEEL_STRUCTURES_CSA":
            clientObject.steel_structure_csa_determine_tau_b = parameters[0].name
            clientObject.factor_of_axial_stiffness = parameters[1]
            clientObject.factor_of_bending_y_stiffness = parameters[2]
            clientObject.factor_of_bending_z_stiffness = parameters[3]
            clientObject.steel_structure_csa_stiffness_factor_of_shear_y_stiffness = parameters[4]
            clientObject.steel_structure_csa_stiffness_factor_of_shear_z_stiffness = parameters[5]
            clientObject.steel_structure_csa_stiffness_factor_of_torsion_stiffness = parameters[6]

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Member Stiffness Modification to client model
        model.clientModel.service.set_member_stiffness_modification(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes
from RSTAB.enums import DesignSituationType

class DesignSituation():

    def __init__(self,
                 no: int = 1,
                 design_situation_type = DesignSituationType.DESIGN_SITUATION_TYPE_A_ACCIDENTAL,
                 active: bool = True,
                 name = None,
                 comment: str = '',
                 params: dict = None,
                 model = Model):

        """
        Args:
            no (int): Design Situation Tag
            design_situation_type (enum): Design Situation Type Enumeration
            active (bool): Enable/Disable Design Situation Activity
            name (str, optional): User-Defined Name
            comment (str, optional): Comments
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client model | Design Situation
        clientObject = model.clientModel.factory.create('ns0:design_situation')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Design Situation No.
        clientObject.no = no

        # Design Situation Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Design Situation Active
        clientObject.active = active

        # Design Situation Type
        clientObject.design_situation_type = design_situation_type.name

        # Design Situation Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Design Situation to client model
        model.clientModel.service.set_design_situation(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString
from RSTAB.enums import AluminumEffectiveLengthsDeterminationMcrEurope

class AluminumEffectiveLengths():
    def __init__(self,
                 no: int = 1,
                 members: str = '1',
                 member_sets: str = '',
                 flexural_buckling_about_y: bool = True,
                 flexural_buckling_about_z: bool = True,
                 torsional_buckling: bool = True,
                 lateral_torsional_buckling: bool = True,
                 principal_section_axes: bool = True,
                 geometric_section_axes: bool = True,
                 name: str = 'SEL1',
                 intermediate_nodes: bool = False,
                 different_properties: bool = True,
                 factors_definition_absolute: bool = False,
                 import_from_stability_analysis_enabled: bool = False,
                 determination_of_mcr = AluminumEffectiveLengthsDeterminationMcrEurope.DETERMINATION_EUROPE_EIGENVALUE,
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Aluminum Effective Length Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            flexural_buckling_about_y (bool): Flexural Buckling About Y Option
            flexural_buckling_about_z (bool): Flexural Buckling About Z Option
            torsional_buckling (bool): Torsional Buckling Option
            lateral_torsional_buckling (bool): Lateral Torsional Buckling Option
            principal_section_axes (bool): Principal Section Axes Option
            geometric_section_axes (bool): Geometric Section Axes Option
            name (str): User Defined Effective Length Name
            intermediate_nodes (bool): Intermediate Nodes Option
            different_properties (bool): Different Properties Option
            factors_definition_absolute (bool): Absolute Factors Definition Option
            import_from_stability_analysis_enabled (bool): Import From Stability Analysis Option
            determination_of_mcr (enum): Aluminum Effective Lengths Determination Mcr Europe Enumeration Item
            comment (str, optional): Comment
            params (dict, optional): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Types For Aluminum Design Effective Lengths
        clientObject = model.clientModel.factory.create('ns0:aluminum_effective_lengths')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Effective Lengths No.
        clientObject.no = no

        # Effective Lengths Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Effective Lengths Assigned Member Sets
        clientObject.member_sets = ConvertToDlString(member_sets)

        # Effective Lengths Flexural Buckling About Y
        clientObject.flexural_buckling_about_y = flexural_buckling_about_y

        # Effective Lengths Flexural Buckling About Z
        clientObject.flexural_buckling_about_z = flexural_buckling_about_z

        # Effective Lengths Torsional Buckling
        clientObject.torsional_buckling = torsional_buckling

        # Effective Lengths Lateral Torsional Buckling
        clientObject.lateral_torsional_buckling = lateral_torsional_buckling

        # Effective Lengths Principal Section Axes
        clientObject.principal_section_axes = principal_section_axes

        # Effective Lengths Geometric Section Axes
        clientObject.geometric_section_axes = geometric_section_axes

        # Effective Lengths User Defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Effective Lengths Intermediate Points
        clientObject.intermediate_nodes = intermediate_nodes

        # Effective Lengths Different Properties
        clientObject.different_properties = different_properties

        # Effective Lengths Factors Definition Absolute
        clientObject.factors_definition_absolute = factors_definition_absolute

        # Effective Lengths Import From Stability Analysis Enabled
        clientObject.import_from_stability_analysis_enabled = import_from_stability_analysis_enabled

        # Effective Lengths Determination MCR Europe
        clientObject.determination_mcr_europe = determination_of_mcr.name

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Aluminum Effective Lengths to client model
        model.clientModel.service.set_aluminum_effective_lengths(clientObject)



from RSTAB.initModel import Model, clearAttributes, deleteEmptyAttributes, ConvertToDlString
from RSTAB.enums import AluminumMemberLocalSectionReductionType, MultipleOffsetDefinitionType, FastenerDefinitionType

class AluminumMemberLocalSectionReduction():

    def __init__(self,
                 no: int = 1,
                 members: str = '1',
                 member_sets: str = '',
                 components: list = [
                    [AluminumMemberLocalSectionReductionType.REDUCTION_COMPONENT_TYPE_DESIGN_PARAMETERS, 1.0, False,\
                     FastenerDefinitionType.DEFINITION_TYPE_ABSOLUTE, 0.5, 2,\
                     MultipleOffsetDefinitionType.OFFSET_DEFINITION_TYPE_ABSOLUTE, 1.0]
                                    ],
                 name: str = '',
                 comment: str = '',
                 params: dict = None,
                 model = Model):
        """
        Args:
            no (int): Member Local Section Reduction Tag
            members (str): Assigned Members
            member_sets (str): Assigned Member Sets
            components (list of lists): Components Table Definition
                components[i][0] (enum): Aluminum Member Local Section Reduction Type Enumeration
                components[i][1] (float): Position Value
                components[i][2] (bool): Enable/Disable Multiple Option
                components[i][3] (enum): Fastener Definition Type Enumeration
                for components[i][3] == FastenerDefinitionType.DEFINITION_TYPE_ABSOLUTE
                    components[i][4] (float): Reduction Area
                for components[i][3] == FastenerDefinitionType.DEFINITION_TYPE_RELATIVE
                    components[i][4] (float): Reduction Area Factor (value must be between 0.0 and 1.0)
                if components[i][2] == True
                    components[i][5] (int): Multiple Number
                    components[i][6] (enum): Multiple Offset Definition Type Enumeration
                    for MultipleOffsetDefinitionType.OFFSET_DEFINITION_TYPE_ABSOLUTE
                        components[i][7] (float): Multiple Offset Value
                    for MultipleOffsetDefinitionType.OFFSET_DEFINITION_TYPE_RELATIVE
                        components[i][7] (float): Multiple Offset Value (value must be between 0.0 and 1.0)
            name (str): User Defined  Member Local Section Reduction Name
            comment (str): Comments
            params (dict): Any WS Parameter relevant to the object and its value in form of a dictionary
            model (RSTAB Class, optional): Model to be edited
        """

        # Client Model | Types For Aluminum Member Local Section Reduction
        clientObject = model.clientModel.factory.create('ns0:aluminum_member_local_section_reduction')

        # Clears object atributes | Sets all atributes to None
        clearAttributes(clientObject)

        # Local Section Reduction No.
        clientObject.no = no

        # Local Section Reduction Assigned Members
        clientObject.members = ConvertToDlString(members)

        # Local Section Reduction Assigned Member Sets
        clientObject.member_sets = ConvertToDlString(member_sets)

        # Local Section Reduction User defined Name
        if name:
            clientObject.user_defined_name_enabled = True
            clientObject.name = name

        # Local Section Reduction Components
        clientObject.components = model.clientModel.factory.create('ns0:array_of_aluminum_member_local_section_reduction_components')

        for i,j in enumerate(components):
            smlsr = model.clientModel.factory.create('ns0:aluminum_member_local_section_reduction_components_row')
            smlsr.no = i+1
            smlsr.row = model.clientModel.factory.create('ns0:aluminum_member_local_section_reduction_components')
            clearAttributes(smlsr.row)
            smlsr.row.reduction_type = components[i][0].name
            smlsr.row.position = components[i][1]
            smlsr.row.multiple = components[i][2]
            smlsr.row.fastener_definition_type = components[i][3].name
            if smlsr.row.fastener_definition_type == "DEFINITION_TYPE_ABSOLUTE":
                smlsr.row.reduction_area = components[i][4]
            elif smlsr.row.fastener_definition_type == "DEFINITION_TYPE_RELATIVE":
                smlsr.row.reduction_area_factor = components[i][4]
            if smlsr.row.multiple:
                smlsr.row.multiple_number = components[i][5]
                smlsr.row.multiple_offset_definition_type = components[i][6].name
                smlsr.row.multiple_offset = components[i][7]
            else:
                smlsr.row.multiple_offset_definition_type = None # important

            clientObject.components.aluminum_member_local_section_reduction_components.append(smlsr)

        # Comment
        clientObject.comment = comment

        # Adding optional parameters via dictionary
        if params:
            for key in params:
                clientObject[key] = params[key]

        # Delete None attributes for improved performance
        deleteEmptyAttributes(clientObject)

        # Add Aluminum Member Local Section Reduction to Client Model
        model.clientModel.service.set_aluminum_member_local_section_reduction(clientObject)
